/*  HasciiCam 0.7
 *  (c) 2000 Denis Roio aka jaromil
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <pwd.h>

#include <linux/types.h>
#include <linux/videodev.h>
#include <jpeglib.h>
#include <aalib.h>

#include "ftp.h"

unsigned char *image;

/* line command stuff */

char *version =
"HasciiCam v0.7 - (h)ascii 4 the masses!\n"
"(c)2000 Denis Roio aka jaromil <jaromil@dyne.org>\n"
"[ http://ascii.dyne.org ]\n\n";

char *help =
/* "\x1B" "c" <--- SCREEN CLEANING ESCAPE CODE
   why here? just a reminder */
"Usage: hasciicam [options]\n"
"options:\n"
"-h --help         this help\n"
"-v --version      version information\n"
"-q --quiet        be quiet\n"
"-m --mode         mode: live|html|text      - default live\n"
"-d --device       video grabbing device     - default /dev/video\n"
"-i --input        input: svideo|composite   - default composite\n"
"-n --norm         norm: pal|ntsc|secam|auto - default auto\n"
"-s --size         ascii image size WxH      - default 96x72\n"
"-r --refresh      refresh delay             - default 2\n"
"-o --aafile       dumped file               - default hasciicam.[txt|html]\n"
"-f --ftp          ie: user@ftp.host.it:/dir - default none\n"
"-b --aabright     ascii brightness          - default 50\n"  
"-c --aacontrast   ascii contrast            - default 10\n"
"-g --aagamma      ascii gamma               - default 10\n"
"-I --invert       invert colors             - default off\n"
"-B --background   background color (#hex)   - default #000000\n"
"-F --foreground   foreground color (#hex)   - default #00FF00\n"
"-O --jpgfile      jpeg file                 - default none\n"
"-T --timestamp    put timestamp over jpeg   - default off\n"
"-Q --jpgqual      jpeg rendering quality    - default 65\n"
//"-B --bttvbright   capture device brightness\n"
//"-C --bttvcontrast capture device contrast\n"
//"-G --bttvgamma    capture device gamma\n"
"\n"
"--\n"
"nifty uses:\n"
"./hasciicam -m html -o watchme.html -f user@ftp.foo.net:/home/user/www\n"
"        puts your ascii into your homepage!\n"
"./hasciicam -s79x24 -m text -o ~/.plan\n"
"        puts your ascii into your .plan! (fingercam)\n";

static const struct option long_options[] = {
  { "help", no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'v' },
  { "quiet", no_argument, NULL, 'q' },
  { "mode", required_argument, NULL, 'm' },
  { "device", required_argument, NULL, 'd'},
  { "input", required_argument, NULL, 'i'},
  { "norm", required_argument, NULL, 'n'},
  { "size", required_argument, NULL, 's' },
  { "refresh", required_argument, NULL, 'r' },
  { "aafile", required_argument, NULL, 'o' },
  { "ftp", required_argument, NULL, 'f' },
  { "aabright", required_argument, NULL, 'b' },
  { "aacontrast", required_argument, NULL, 'c' },
  { "aagamma", required_argument, NULL, 'g' },
  { "invert", no_argument, NULL, 'I'},
  { "background", required_argument, NULL, 'B'},
  { "foreground", required_argument, NULL, 'F'},
  { "jpgfile", required_argument, NULL, 'O' },
  { "timestamp", no_argument, NULL, 'T' },
  { "jpgqual", required_argument, NULL, 'Q' },
  //  { "bttvbright", required_argument, NULL, 'B' },
  //  { "bttvcontrast", required_argument, NULL, 'C' },
  //  { "bttvgamma", required_argument, NULL, 'G' },
  {0, 0, 0, 0}
};

char *short_options = "hvqm:d:i:n:s:r:o:f:b:c:g:IB:F:O:Q:";

/* default configuration */
int quiet = 0;
int mode = 0;
int useftp = 0;
char device[] = "/dev/video\0                                                                 ";
int input = 1;
int norm = VIDEO_MODE_AUTO;
int width = 96;
int height = 72;
int live = 0;
int refresh = 2;
char aafile[] = "\0                                                                           ";
char ftp[160];
char ftp_user[32];
char ftp_host[64];
char ftp_dir[64];
char *ftp_pass;

char jpgfile[] = "\0                                                                           ";
int timestamp = 0;
int jpgqual = 65;
int aabright = 50;
int aacontrast = 10;
int aagamma = 10;
int invert = 0;
char *background = "#000000";
char *foreground = "#00FF00";

/* if we're grabbing a 24bit image or 8bit greyscale */
int grab24 = 0;

/* if width&height have been manually changed */
int whchanged = 0;

/* ascii context & html formatting stuff*/
aa_context *ascii_img;

char html_header[512];

static char *html_escapes[] = {"<", "&lt;", ">", "&gt;", "&", "&amp;", NULL};

struct aa_format aa_html_format =
{
  79, 25,
  0, 0,
  0,
  AA_NORMAL_MASK | AA_BOLD_MASK | AA_BOLDFONT_MASK,
  NULL,
  "Pure html",
  ".html",
  html_header,
  "</PRE>\n</FONT>\n</BODY>\n</HTML>\n",
  "\n",
  /*The order is:normal, dim, bold, boldfont, reverse, special*/
  { "%s", "%s", "%s", "%s", "%s", },
  {"", "", "<B>", "", "<B>" },
  {"", "", "</B>", "", "</B>" },
  html_escapes
};

/* jpeg */

void
swap_rgb24(char *mem, int n)
{
    char  c;
    char *p = mem;
    int   i = n;
    
    while (--i) {
	c = p[0]; p[0] = p[2]; p[2] = c;
	p += 3;
    }
}

void
resample_8bit(char *mem, int n)
{
  char r,g,b;
  int c,cc;

  c=0;

  for(cc=0;cc<n;cc++) {
    r = mem[c];
    g = mem[c+1];
    b = mem[c+2];
    mem[cc] = (r+g+b)/3;
    c+=3;
  }
}

# define CHAR_HEIGHT  11
# define CHAR_WIDTH   6
# define CHAR_START   4
# include "font_6x11.h"

void
add_text(char *image, int w, int h)
{
    time_t      t;
    struct tm  *tm;
    char        line[128],*ptr;
    int         i,x,y,f,len;

    time(&t);
    tm = localtime(&t);
    len = strftime(line,127,"%d.%m.%Y %H:%M:%S",tm);
    fprintf(stderr,"%s\n",line);

    for (y = 0; y < CHAR_HEIGHT; y++) {
	ptr = image + 3 * w * (h-CHAR_HEIGHT-2+y) + 12;
	for (x = 0; x < len; x++) {
	    f = fontdata[line[x] * CHAR_HEIGHT + y];
	    for (i = CHAR_WIDTH-1; i >= 0; i--) {
		if (f & (CHAR_START << i)) {
		    ptr[0] = 255;
		    ptr[1] = 255;
		    ptr[2] = 255;
		}
		ptr += 3;
	    }
	}
    }
}

int
write_jpeg(char *filename, char *data, int w, int h)
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    FILE *fp;
    int i;
    unsigned char *line;

    if (NULL == (fp = fopen(filename,"w"))) {
	fprintf(stderr,"can't open %s for writing: %s\n",
		filename,strerror(errno));
	return -1;
    }

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    jpeg_stdio_dest(&cinfo, fp);
    cinfo.image_width = w;
    cinfo.image_height = h;
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_RGB;
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, jpgqual, TRUE);
    jpeg_start_compress(&cinfo, TRUE);

    for (i = 0, line = data; i < h; i++, line += w*3)
	jpeg_write_scanlines(&cinfo, &line, 1);
    
    jpeg_finish_compress(&(cinfo));
    jpeg_destroy_compress(&(cinfo));
    fclose(fp);

    return 0;
}

/* v4l */

static struct video_capability grab_cap;
static struct video_mmap grab_buf;
static struct video_channel grab_chan;
struct video_picture grab_pic;
static int grab_fd = -1;
static int grab_size;

void grab_detect() {
  int counter, res;
  char *capabilities[] = {
    "  VID_TYPE_CAPTURE          can capture to memory",
    "  VID_TYPE_TUNER            has a tuner of some form",
    "  VID_TYPE_TELETEXT         has teletext capability",
    "  VID_TYPE_OVERLAY          can overlay its image onto the frame buffer",
    "  VID_TYPE_CHROMAKEY        overlay is chromakeyed",
    "  VID_TYPE_CLIPPING         overlay clipping supported",
    "  VID_TYPE_FRAMERAM         overlay overwrites frame buffer memory",
    "  VID_TYPE_SCALES           supports image scaling",
    "  VID_TYPE_MONOCHROME       image capture is grey scale only",
    "  VID_TYPE_SUBCAPTURE       capture can be of only part of the image"
  };
  
  res = ioctl(grab_fd,VIDIOCGCAP,&grab_cap);
  if(res<0) {
    perror("Error in VIDIOCGCAP ");
    exit(1);
  }
  
  fprintf(stderr,"-- detecting hardware\n");
  fprintf(stderr,"Device used is %s\n",device);
  fprintf(stderr,"%s\n",grab_cap.name);
  fprintf(stderr,"%u channels detected\n",grab_cap.channels);
  fprintf(stderr,"maximum capture size is %ux%u\n",grab_cap.maxwidth,grab_cap.maxheight);
  fprintf(stderr,"minimum capture size is %ux%u\n",grab_cap.minwidth,grab_cap.minheight);
  fprintf(stderr,"Video capabilities:\n");
  for (counter=0;counter<10;counter++)
    if (grab_cap.type & (1 << counter)) fprintf(stderr,"%s\n",capabilities[counter]);
  if((grab_pic.palette|VIDEO_PALETTE_GREY)&VIDEO_PALETTE_GREY)
    fprintf(stderr,"  VIDEO_PALETTE_GREY        device is able to grab greyscale frames\n");
  else {
    grab24 = 1;
  }

  if(strncmp(grab_cap.name,"Buz",3)==0)
    /* the device is a iomega Buz. We need to check that 'cause the iomegaBuz
       claims to have the VIDEO_PALETTE_GREY capability, but then grabs 24bit
       images, overflowing the buffer. So we don't trust the capability report
       and turn on software 24bit->8bit downsampling */
    grab24 = 1;

  fprintf(stderr,"--\n");
}

unsigned char *grab_init() {
  unsigned char *grab_data;
    if (-1 == (grab_fd = open(device,O_RDWR))) {
	perror("Error in opening video capture device");
	exit(1);
    }

    if (quiet==0) grab_detect();

    /* set image source and TV norm */
    grab_chan.channel = input;
    if (-1 == ioctl(grab_fd,VIDIOCGCHAN,&grab_chan)) {
	perror("ioctl VIDIOCGCHAN ");
	exit(1);
    }

    if (-1 == ioctl(grab_fd,VIDIOCSCHAN,&grab_chan)) {
      perror("ioctl VIDIOCSCHAN ");
      exit(1);
    }

    if (-1 == ioctl(grab_fd, VIDIOCGPICT, &grab_pic)) {
      perror("ioctl VIDIOCGPICT ");
      exit(1);
    }

    if(grab24==0)
      grab_buf.format = VIDEO_PALETTE_GREY;
    else
      grab_buf.format = VIDEO_PALETTE_RGB24;
    
    grab_buf.frame  = 0;
    grab_buf.height = height*2;
    grab_buf.width = width*2;
    grab_size = grab_buf.width * grab_buf.height * 3;
    grab_data = mmap(0,grab_size,PROT_READ|PROT_WRITE,MAP_SHARED,grab_fd,0);
    if (-1 == (int)grab_data) {
	perror("Cannot allocate video surface ");
	exit(1);
    }

    /* init the html header */
    snprintf(&html_header[0],512,"<HTML>\n <HEAD> <TITLE>!!Ascii Cam Streaming!!</TITLE>\n<META HTTP-EQUIV=\"refresh\" CONTENT=\"%u\"; url=\"%s\">\n<META HTTP-EQUIV=\"Pragma\" CONTENT=\"no-cache\">\n<STYLE TYPE=\"text/css\">\n<!--\npre {\nletter-spacing: 1px;\nlayer-background-color: Black;\nleft : auto;\nline-height : 5px;\n}\n-->\n</STYLE>\n</HEAD>\n<BODY bgcolor=%s text=%s>\n<FONT SIZE=1>\n<PRE>\n",refresh,aafile,background,foreground);
    return(grab_data);
}

int grab_one(int width, int height, int palette)
{
  if(grab24==0) {
    grab_buf.format = palette;
    if (-1 == ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buf)) {
      perror("ioctl VIDIOCMCAPTURE");
    } else {
      if (-1 == ioctl(grab_fd,VIDIOCSYNC,&grab_buf)) {
	perror("ioctl VIDIOCSYNC");
      } else {
	width  = grab_buf.width;
	height = grab_buf.height;
	return 1;
      }
    }
  } else {
    grab_buf.format = VIDEO_PALETTE_RGB24;
    if (-1 == ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buf)) {
      perror("ioctl VIDIOCMCAPTURE");
    } else {
      if (-1 == ioctl(grab_fd,VIDIOCSYNC,&grab_buf)) {
	perror("ioctl VIDIOCSYNC");
      } else {
	//	swap_rgb24(image,grab_buf.width*grab_buf.height);
	resample_8bit(image,grab_buf.width*grab_buf.height);
	width = grab_buf.width;
	height = grab_buf.height;
	return(1);
      }
    }
  }
  return -1;
}

/* misc */

void goout() {
  aa_close(ascii_img);
  if(grab_fd>0) close(grab_fd);
}

void config_init(int argc, char *argv[]) {
  int res;
  
  do {
    res = getopt_long(argc, argv, short_options, long_options, NULL);
      switch(res) {
      case 'h':
	fprintf(stderr,"%s",help);
	exit(0);
	break;
      case 'v':
	fprintf(stderr,"%s",version);
	exit(0);
	break;
      case 'q':
	quiet = 1;
	break;
      case 'm':
	if(strcmp(optarg,"live") == 0) {
	  if(useftp) {
	    fprintf(stderr,"heek, ftp option makes no sense with live mode!\n");
	    useftp = 0;
	  }
	  mode = 0;
	} else if (strcmp(optarg,"html") == 0) {
	  mode = 1;
	  if(aafile[0]=='\0')
	    strcpy(aafile,"hasciicam.html");
	} else if (strcmp(optarg,"text") == 0) {
	  mode = 2;
	  if(aafile[0]=='\0')
	    strcpy(aafile,"hasciicam.asc");
	} else
	  fprintf(stderr,"!! invalid mode selected, using live\n");
	break;
      case 'd':
	strncpy(device,optarg,83);
	break;
      case 'i':
	if(strcmp(optarg,"svideo") == 0)
	  input = 2;
	else if(strcmp(optarg,"composite") == 0)
	  input = 1;
	else
	  fprintf(stderr,"!! invalid input source selected, using composite\n");
	break;
      case 'n':
	if(strcmp(optarg,"pal") == 0)
	  norm = VIDEO_MODE_PAL;
	else if(strcmp(optarg,"ntsc") == 0)
	  norm = VIDEO_MODE_NTSC;
	else if(strcmp(optarg,"secam") == 0)
	  norm = VIDEO_MODE_SECAM;
	else if(strcmp(optarg,"auto") == 0)
	  norm = VIDEO_MODE_AUTO;
	else
	  fprintf(stderr,"!! invalid video norm selected, using auto\n");
	break;
      case 's':
	{
	  char *t;
	  char *tt;
	  t = optarg;
	  while(isdigit(*t)) t++;
	  *t = 0;
	  width = atoi(optarg);
	  tt = ++t;
	  while(isdigit(*tt)) tt++;
	  *tt = 0;
	  height = atoi(t);
	  whchanged = 1;
	}
	break;
      case 'r':
	refresh = atoi(optarg);
	break;
      case 'o':
	strncpy(aafile,optarg,83);
	break;
      case 'f':
	if(mode!=0) {
	  strncpy(ftp,optarg,160);
	  useftp = 1;
	} else
	  fprintf(stderr,"heek, ftp option makes no sense with live mode!\n");
	break;
      case 'b':
	aabright = atoi(optarg);
	break;
      case 'c':
	aacontrast = atoi(optarg);
	break;
      case 'g':
	aagamma = atoi(optarg);
	break;
      case 'I':
	invert = 1;
	break;
      case 'B':
	strncpy(background,optarg,7);
	break;
      case 'F':
	strncpy(foreground,optarg,7);
	break;
      case 'O':
	strncpy(jpgfile,optarg,83);
	break;
      case 'T':
	if(strcmp(jpgfile,"") != 0)
	  timestamp = 1;
	else
	  fprintf(stderr,"!! setting timestamp without enabling jpeg makes no sense\n");
	break;
      case 'Q':
	if(strcmp(jpgfile,"") != 0)
	  jpgqual = atoi(optarg);
	else
	  fprintf(stderr,"!! setting jpeg quality without enabling jpeg makes no sense\n");
	break;
	/*
	  case 'B':
	  bttvbright = atoi(optarg);
	  break;
	  case 'C':
	  bttvcontrast = atoi(optarg);
	  break;
	  case 'G':
	  bttvgamma = atoi(optarg);
	  break;
	*/
      }
  } while(res > 0);

  if((width==0)||(height==0)) {
    printf("Error: invalid size %dx%d\n",width,height);
    exit(0);
  }

  if(useftp) {
    char *p,*pp;
    p = ftp;
    while(*p!='@') {
      if((ftp-p)<32)
	p++; 
      else {
	printf("Error: malformed ftp command: %s\n",ftp);
	exit(0);
      }
    }
    *p = '\0';
    strcpy(ftp_user,ftp);
    p++; pp = p;
    while(*p!=':') {
      if((pp-p)<64)
	p++; 
      else {
	printf("Error: malformed ftp command: %s\n",ftp);
	exit(0);
      }
    }
    *p = '\0';
    strcpy(ftp_host,pp);
    p++; pp = p;
    while(*p!='\0') {
      if((pp-p)<64)
	p++; 
      else {
	printf("Error: malformed ftp command: %s\n",ftp);
	exit(0);
      }
    }
    strcpy(ftp_dir,pp);
  }
}

/* here we go */

int
main(int argc, char *argv[])
{
  /* !!! grabbing height & width should be double the ascii context width and height !!!*/

  int awidth, aheight;
  struct aa_renderparams ascii_parms;
  struct aa_hardware_params ascii_surf;
  struct aa_savedata ascii_save;

  fprintf(stderr,"%s",version);

  /* gathering line commands */
  config_init(argc, argv);
  
  /* initialize aalib default params */
  memcpy(&ascii_surf,&aa_defparams,sizeof(struct aa_hardware_params));

  /* sets the default size on live mode */
  if(whchanged==0&&mode==0) {
    width = 80;
    height = 40;
  } 
  
  ascii_surf.width = width;
  ascii_surf.height = height;  

  /* bttv grabber init */
  image = grab_init();
  
  switch(mode) {
  case 0:
    fprintf(stderr,"using LIVE mode\n");
    break;
  case 1:
    ascii_save.name = aafile;
    ascii_save.format = &aa_html_format;
    ascii_save.file = NULL;
    fprintf(stderr,"using HTML mode dumping to file %s",aafile);
    if(useftp)
      fprintf(stderr," ftp-pushing on %s%s\n",ftp_host,ftp_dir);
    else
      fprintf(stderr,"\n");
    break;
  case 2:
    ascii_save.name = aafile;
    ascii_save.format = &aa_text_format;
    ascii_save.file = NULL;
    fprintf(stderr,"using TEXT mode dumping to file %s",aafile);
    if(useftp)
      fprintf(stderr," ftp-pushing on ftp://%s%s\n",ftp_host,ftp_dir);
    else
      fprintf(stderr,"\n");
    break;
  default:
    break;
  }
  
  /* aalib init */
  if(mode>0)
    ascii_img = aa_init(&save_d,&ascii_surf,&ascii_save);
  else
    ascii_img = aa_autoinit(&ascii_surf);

  /* ftp init */
  if(useftp) {
    char temp[160];
    ftp_init(0,0);
    sprintf(temp,"password for %s@%s : ",ftp_user,ftp_host);
    ftp_pass = getpass(temp);
    ftp_connect(ftp_host,ftp_user,ftp_pass,ftp_dir);
  }
  
  /* be sure to exit clean */
  atexit(goout);

  awidth = aa_scrwidth(ascii_img);
  aheight = aa_scrheight(ascii_img);
  fprintf(stderr,"grabbed image is %dx%d - ",grab_buf.width,grab_buf.height);
  fprintf(stderr,"ascii context is %dx%d\n",awidth,aheight);
  
  /* setting ascii rendering parameters */
  ascii_parms.bright = aabright;
  ascii_parms.contrast = aacontrast;
  ascii_parms.gamma = aagamma;
  ascii_parms.dither = AA_FLOYD_S;
  ascii_parms.inversion = invert;
  ascii_parms.randomval = 0;

  /* cycle until ctrl-c */
  for(;;) {
    grab_one(width,height,VIDEO_PALETTE_GREY);

    memcpy(aa_image(ascii_img),image,awidth*aheight*4);

    aa_render(ascii_img, &ascii_parms, 0, 0, awidth, aheight);
    
    if (strcmp(jpgfile,"") != 0) {
      grab_one(width,height,VIDEO_PALETTE_RGB24);
      swap_rgb24(image,grab_buf.width*grab_buf.height);
      if(timestamp) add_text(image,width<<1,height<<1);
      write_jpeg(jpgfile,image,width<<1,height<<1);
    }

    aa_flush(ascii_img);

    if(useftp) {
      if(!ftp_connected)
	ftp_connect(ftp_host,ftp_user,ftp_pass,ftp_dir);

      /* scolopendro is the tmp file being renamed */
      ftp_upload(aafile,aafile,"scolopendro");
    }

    if(mode>0) sleep(refresh);
  }

  exit(0);
}
